from .block import Block
from .overloaded_type import OverloadedType, register_overloaded_type, create_overloaded_object
from .tape import get_working_tape, annotate_tape, stop_annotating


def annotate_operator(operator):
    """Decorate float operator like __add__, __sub__, etc.

    The provided operator is only expected to create the Block that
    corresponds to this operation. The decorator returns a wrapper code
    that checks whether annotation is needed, ensures all arguments are
    overloaded, calls the block-creating operator, and puts the Block
    on tape."""

    # the actual float operation is derived from the name of operator
    try:
        float_op = getattr(float, operator.__name__)
    except AttributeError:
        if operator.__name__ == '__div__':
            float_op = float.__truediv__
        else:
            raise

    def annotated_operator(self, *args):
        output = float_op(self, *args)
        if output is NotImplemented:
            return NotImplemented

        # ensure all arguments are of OverloadedType
        args = [arg if isinstance(arg, OverloadedType) else self.__class__(arg) for arg in args]

        output = self.__class__(output)
        if annotate_tape():
            block = operator(self, *args)

            tape = get_working_tape()
            tape.add_block(block)
            block.add_output(output.block_variable)

        return output

    return annotated_operator


@register_overloaded_type
class AdjFloat(OverloadedType, float):
    def __new__(cls, *args, **kwargs):
        return float.__new__(cls, *args)

    def __init__(self, *args, **kwargs):
        super(AdjFloat, self).__init__(*args, **kwargs)

    @annotate_operator
    def __mul__(self, other):
        return MulBlock(self, other)

    @annotate_operator
    def __div__(self, other):
        return DivBlock(self, other)

    @annotate_operator
    def __truediv__(self, other):
        return DivBlock(self, other)

    @annotate_operator
    def __neg__(self):
        return NegBlock(self)

    @annotate_operator
    def __rmul__(self, other):
        return MulBlock(self, other)

    @annotate_operator
    def __add__(self, other):
        return AddBlock(self, other)

    @annotate_operator
    def __radd__(self, other):
        return AddBlock(self, other)

    @annotate_operator
    def __sub__(self, other):
        return SubBlock(self, other)

    @annotate_operator
    def __rsub__(self, other):
        # NOTE: order is important here
        return SubBlock(other, self)

    @annotate_operator
    def __pow__(self, power):
        return PowBlock(self, power)

    def _ad_convert_type(self, value, options={}):
        return AdjFloat(value)

    def _ad_create_checkpoint(self):
        # Floats are immutable.
        return self

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    def _ad_mul(self, other):
        return self * other

    def _ad_add(self, other):
        return self + other

    def _ad_dot(self, other):
        return float.__mul__(self, other)

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        dst = type(dst)(src[offset:offset + 1])
        offset += 1
        return dst, offset

    @staticmethod
    def _ad_to_list(value):
        return [value]

    def _ad_copy(self):
        return self


_min = min
_max = max


def min(a, b, **kwargs):
    annotate = annotate_tape(kwargs)
    if annotate:
        # Ensure a and b are of OverloadedType
        a = create_overloaded_object(a)
        b = create_overloaded_object(b)

        block = MinBlock(a, b)
        tape = get_working_tape()
        tape.add_block(block)

    with stop_annotating():
        out = _min(a, b)
    out = AdjFloat(out)

    if annotate:
        block.add_output(out.block_variable)
    return out


def max(a, b, **kwargs):
    annotate = annotate_tape(kwargs)
    if annotate:
        # Ensure a and b are of OverloadedType
        a = create_overloaded_object(a)
        b = create_overloaded_object(b)

        block = MaxBlock(a, b)
        tape = get_working_tape()
        tape.add_block(block)

    with stop_annotating():
        out = _max(a, b)
    out = AdjFloat(out)

    if annotate:
        block.add_output(out.block_variable)
    return out


class MinBlock(Block):
    def __init__(self, a, b):
        super().__init__()
        self.add_dependency(a)
        self.add_dependency(b)

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        adj_input = adj_inputs[0]
        active_idx = 0 if inputs[0] <= inputs[1] else 1
        if idx == active_idx:
            return adj_input
        else:
            return 0.

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        idx = 0 if inputs[0] <= inputs[1] else 1
        return tlm_inputs[idx]

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        return self.evaluate_adj_component(inputs, hessian_inputs, block_variable, idx, prepared)

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return _min(inputs[0], inputs[1])


class MaxBlock(Block):
    def __init__(self, a, b):
        super().__init__()
        self.add_dependency(a)
        self.add_dependency(b)

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        adj_input = adj_inputs[0]
        active_idx = 0 if inputs[0] >= inputs[1] else 1
        if idx == active_idx:
            return adj_input
        else:
            return 0.

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        idx = 0 if inputs[0] >= inputs[1] else 1
        return tlm_inputs[idx]

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        return self.evaluate_adj_component(inputs, hessian_inputs, block_variable, idx, prepared)

    def recompute_component(self, inputs, block_variable, idx, prepared):
        r = _max(inputs[0], inputs[1])
        return r


class FloatOperatorBlock(Block):
    # the float operator annotated in this Block
    operator = None
    symbol = None

    def __init__(self, *args):
        super(FloatOperatorBlock, self).__init__()
        # the terms are stored seperately here and added as dependencies
        # this is because get_dependencies() only returns the terms with
        # duplicates taken out; for evaluation however order and position
        # of the terms is significant
        self.terms = [arg.block_variable for arg in args]
        for dep in args:
            self.add_dependency(dep)

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return self.operator(*(term.saved_output for term in self.terms))

    def __str__(self):
        return f"{self.terms[0]} {self.symbol} {self.terms[1]}"


class PowBlock(FloatOperatorBlock):
    operator = staticmethod(float.__pow__)
    symbol = "**"

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        base_value = inputs[0]
        exponent_value = inputs[1]
        adj_input = adj_inputs[0]

        if idx == 0:
            return float.__mul__(float.__mul__(adj_input, exponent_value),
                                 float.__pow__(base_value, exponent_value - 1))
        else:
            from numpy import log
            return float.__mul__(float.__mul__(adj_input, log(base_value)),
                                 float.__pow__(base_value, exponent_value))

    def evaluate_tlm(self, markings=False):
        output = self.get_outputs()[0]

        base = self.terms[0]
        exponent = self.terms[1]

        base_value = base.saved_output
        exponent_value = exponent.saved_output

        if base.tlm_value is not None:
            base_tlm = float.__mul__(float.__mul__(base.tlm_value, exponent_value),
                                     float.__pow__(base_value, exponent_value - 1))
            output.add_tlm_output(base_tlm)

        if exponent.tlm_value is not None:
            from numpy import log
            exponent_adj = float.__mul__(float.__mul__(exponent.tlm_value, log(base_value)),
                                         float.__pow__(base_value, exponent_value))
            output.add_tlm_output(exponent_adj)

    def evaluate_hessian(self, markings=False):
        output = self.get_outputs()[0]
        hessian_input = output.hessian_value
        adj_input = output.adj_value
        if hessian_input is None:
            return

        base = self.terms[0]
        exponent = self.terms[1]

        base_value = base.saved_output
        exponent_value = exponent.saved_output

        # First we do the base hessian (minus the mixed derivative)
        if base.tlm_value is not None:
            second_order = float.__mul__(float.__mul__(
                float.__mul__(adj_input, float.__mul__(exponent_value, exponent_value - 1)),
                float.__pow__(base_value, exponent_value - 2)), base.tlm_value)
            base.add_hessian_output(second_order)

        first_order = float.__mul__(float.__mul__(hessian_input, exponent_value),
                                    float.__pow__(base_value, exponent_value - 1))
        base.add_hessian_output(first_order)

        # Then we do the exponent hessian (minus the mixed derivative)
        from numpy import log
        if exponent.tlm_value is not None:
            second_order = float.__mul__(float.__mul__(float.__mul__(adj_input, float.__pow__(log(base_value), 2)),
                                                       float.__pow__(base_value, exponent_value)), exponent.tlm_value)
            exponent.add_hessian_output(second_order)

        first_order = float.__mul__(float.__mul__(hessian_input, log(base_value)),
                                    float.__pow__(base_value, exponent_value))
        exponent.add_hessian_output(first_order)

        # Lastly we add mixed derivative terms
        mixed = float.__mul__(adj_input, float.__mul__(
            float.__pow__(base_value, exponent_value - 1),
            float.__add__(float.__mul__(exponent_value, log(base_value)), 1)))
        if exponent.tlm_value is not None:
            base.add_hessian_output(float.__mul__(exponent.tlm_value, mixed))
        if base.tlm_value is not None:
            exponent.add_hessian_output(float.__mul__(base.tlm_value, mixed))


class AddBlock(FloatOperatorBlock):
    operator = staticmethod(float.__add__)
    symbol = "+"

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        return adj_inputs[0]

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        tlm_output = 0.
        for term in self.terms:
            tlm_input = term.tlm_value

            if tlm_input is None:
                continue

            tlm_output += tlm_input
        return tlm_output

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        return hessian_inputs[0]


class SubBlock(FloatOperatorBlock):
    operator = staticmethod(float.__sub__)
    symbol = "-"

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        if idx == 0:
            return adj_inputs[0]
        else:
            return float.__neg__(adj_inputs[0])

    def evaluate_tlm(self, markings=False):
        output = self.get_outputs()[0]
        tlm_input_0 = self.terms[0].tlm_value
        if tlm_input_0 is not None:
            output.add_tlm_output(tlm_input_0)
        tlm_input_1 = self.terms[1].tlm_value
        if tlm_input_1 is not None:
            output.add_tlm_output(float.__neg__(tlm_input_1))

    def evaluate_hessian(self, markings=False):
        hessian_input = self.get_outputs()[0].hessian_value
        if hessian_input is None:
            return

        self.terms[0].add_hessian_output(hessian_input)
        self.terms[1].add_hessian_output(float.__neg__(hessian_input))


class MulBlock(FloatOperatorBlock):
    operator = staticmethod(float.__mul__)
    symbol = "*"

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        other_idx = 0 if idx == 1 else 1
        return float.__mul__(adj_inputs[0], inputs[other_idx])

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        tlm_output = 0.
        for i, j in zip((0, 1), (1, 0)):
            tlm_input = self.terms[i].tlm_value

            if tlm_input is None:
                continue

            tlm_output += float.__mul__(tlm_input, self.terms[j].saved_output)
        return tlm_output

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        adj_input = adj_inputs[0]
        hessian_input = hessian_inputs[0]
        other_idx = 0 if idx == 1 else 1
        mixed = 0.0
        for other_idx, bv in relevant_dependencies:
            if other_idx != idx and bv.tlm_value is not None:
                mixed = float.__mul__(adj_input, bv.tlm_value)
        return float.__add__(mixed, float.__mul__(hessian_input, inputs[other_idx]))


class DivBlock(FloatOperatorBlock):
    operator = staticmethod(float.__truediv__)
    symbol = "/"

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        if idx == 0:
            return float.__mul__(
                adj_inputs[0],
                float.__truediv__(1., inputs[1])
            )
        else:
            return float.__mul__(
                adj_inputs[0],
                float.__neg__(float.__truediv__(
                    inputs[0],
                    float.__pow__(inputs[1], 2)
                ))
            )

    def evaluate_tlm(self, markings=False):
        output = self.get_outputs()[0]

        if self.terms[0].tlm_value is not None:
            output.add_tlm_output(float.__mul__(
                self.terms[0].tlm_value,
                float.__truediv__(1., self.terms[1].saved_output)
            ))
        if self.terms[1].tlm_value is not None:
            output.add_tlm_output(float.__mul__(
                self.terms[1].tlm_value,
                float.__neg__(float.__truediv__(
                    self.terms[0].saved_output,
                    float.__pow__(self.terms[1].saved_output, 2)
                ))
            ))

    def evaluate_hessian(self, markings=False):
        output = self.get_outputs()[0]
        hessian_input = output.hessian_value
        adj_input = output.adj_value
        if hessian_input is None:
            return

        numerator = self.terms[0]
        denominator = self.terms[1]

        numerator_value = numerator.saved_output
        denominator_value = denominator.saved_output

        # The function is linear in the numerator
        numerator.add_hessian_output(float.__mul__(
            hessian_input,
            float.__truediv__(1., denominator_value)
        ))

        # Now for the denominator
        denominator.add_hessian_output(float.__mul__(
            hessian_input,
            float.__neg__(float.__truediv__(
                numerator_value,
                float.__pow__(denominator_value, 2.)
            ))
        ))

        if denominator.tlm_value is not None:
            denominator.add_hessian_output(float.__mul__(
                float.__mul__(
                    adj_input,
                    float.__truediv__(
                        float.__mul__(2., numerator_value),
                        float.__pow__(denominator_value, 3)
                    )
                ), denominator.tlm_value))

        # Now for mixed derivative
        mixed = float.__neg__(float.__truediv__(adj_input, float.__pow__(denominator_value, 2)))
        if denominator.tlm_value is not None:
            numerator.add_hessian_output(float.__mul__(denominator.tlm_value, mixed))
        if numerator.tlm_value is not None:
            denominator.add_hessian_output(float.__mul__(numerator.tlm_value, mixed))


class NegBlock(FloatOperatorBlock):
    operator = staticmethod(float.__neg__)
    symbol = "-"

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        return float.__neg__(adj_inputs[0])

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        return float.__neg__(tlm_inputs[0])

    def evaluate_hessian(self, markings=False):
        hessian_input = self.get_outputs()[0].hessian_value
        if hessian_input is None:
            return

        self.terms[0].add_hessian_output(float.__neg__(hessian_input))

    def __str__(self):
        return f"{self.symbol} {self.terms[0]}"



from __future__ import print_function
from .reduced_functional import ReducedFunctional
from .tape import no_annotations, get_working_tape
from .enlisting import Enlist
from .control import Control
from .adjfloat import AdjFloat

import numpy


class ReducedFunctionalNumPy(ReducedFunctional):
    """This class implements the reduced functional for given functional and
    controls based on numpy data structures.

    This "NumPy version" of the pyadjoint.ReducedFunctional is created from
    an existing ReducedFunctional object:
    rf_np = ReducedFunctionalNumPy(rf = rf)
    """

    def __init__(self, functional, controls=None, tape=None):
        if isinstance(functional, AdjFloat):
            functional = ReducedFunctional(functional=functional,
                                           controls=controls,
                                           tape=tape)
        self.rf = functional

    def __getattr__(self, item):
        return getattr(self.rf, item)

    def __call__(self, m_array):
        """An implementation of the reduced functional evaluation
            that accepts the control values as an array of scalars

        """
        m_copies = [control.copy_data() for control in self.controls]
        return self.rf.__call__(self.set_local(m_copies, m_array))

    def set_local(self, m, m_array):
        offset = 0
        for i, control in enumerate(self.controls):
            m[i], offset = control.assign_numpy(m[i], m_array, offset)

        return m

    def get_global(self, m):
        m_global = []
        for i, v in enumerate(Enlist(m)):
            if isinstance(v, Control):
                # TODO: Consider if you want this design.
                m_global += v.fetch_numpy(v.control)
            elif hasattr(v, "_ad_to_list"):
                m_global += v._ad_to_list(v)
            else:
                m_global += self.controls[i].control._ad_to_list(v)
        return numpy.array(m_global, dtype="d")

    @no_annotations
    def derivative(self, m_array=None, forget=True, project=False):
        """ An implementation of the reduced functional derivative evaluation
            that accepts the controls as an array of scalars. If no control values are given,
            the result is derivative at the lastest forward run.
        """

        # In the case that the control values have changed since the last forward run,
        # we first need to rerun the forward model with the new controls to have the
        # correct forward solutions
        # TODO: No good way to check. Is it ok to always assume `m_array` is the same as used last in __call__?
        # if m_array is not None:
        #    self.__call__(m_array)
        dJdm = self.rf.derivative()
        dJdm = Enlist(dJdm)

        m_global = []
        for i, control in enumerate(self.controls):
            # This is a little ugly, but we need to go through the control to get to the OverloadedType.
            # There is no guarantee that dJdm[i] is an OverloadedType and not a backend type.
            m_global += control.fetch_numpy(dJdm[i])

        return numpy.array(m_global, dtype="d")

    @no_annotations
    def hessian(self, m_array, m_dot_array):
        """ An implementation of the reduced functional hessian action evaluation
            that accepts the controls as an array of scalars. If m_array is None,
            the Hessian action at the latest forward run is returned. """
        # Calling derivative is needed, see i.e. examples/stokes-shape-opt
        self.derivative()
        m_copies = [control.copy_data() for control in self.controls]
        Hm = self.rf.hessian(self.set_local(m_copies, m_dot_array))
        Hm = Enlist(Hm)

        m_global = []
        for i, control in enumerate(self.controls):
            # This is a little ugly, but we need to go through the control to get to the OverloadedType.
            # There is no guarantee that dJdm[i] is an OverloadedType and not a backend type.
            m_global += control.fetch_numpy(Hm[i])

        tape = get_working_tape()
        tape.reset_variables()

        return numpy.array(m_global, dtype="d")

    def obj_to_array(self, obj):
        return self.get_global(obj)

    def get_controls(self):
        m = [p.tape_value() for p in self.controls]
        return self.obj_to_array(m)

    def set_controls(self, array):
        m = [p.tape_value() for p in self.controls]
        m = self.set_local(m, array)
        for control, m_i in zip(self.controls, m):
            control.update(m_i)
        return m


def set_local(coeffs, m_array):
    offset = 0
    for m in Enlist(coeffs):
        _, offset = m._ad_assign_numpy(m, m_array, offset)

    return coeffs


def gather(m):
    if isinstance(m, list):
        return list(map(gather, m))
    elif hasattr(m, "_ad_to_list"):
        return m._ad_to_list(m)
    else:
        return m  # Assume it is gathered already



# flake8: noqa

__version__ = '2019.1.2'
__author__  = 'Sebastian Kenji Mitusch'
__credits__ = []
__license__ = 'LGPL-3'
__maintainer__ = 'Sebastian Kenji Mitusch'
__email__ = 'sebastkm@math.uio.no'

from .block import Block
from .tape import (Tape,
                   set_working_tape, get_working_tape, no_annotations,
                   annotate_tape, stop_annotating, pause_annotation, continue_annotation)
from .adjfloat import AdjFloat
from .reduced_functional import ReducedFunctional
from .drivers import compute_gradient, compute_hessian, solve_adjoint
from .verification import taylor_test, taylor_to_dict
from .overloaded_type import OverloadedType, create_overloaded_object
from .control import Control
from .optimization.optimization import minimize, maximize, print_optimization_methods
from .optimization.optimization_problem import MinimizationProblem
from .optimization.ipopt_solver import IPOPTSolver
from .optimization.rol_solver import ROLSolver
from .optimization.constraints import InequalityConstraint, EqualityConstraint
from .optimization.moola_problem import MoolaOptimizationProblem



from .enlisting import Enlist
from .tape import get_working_tape, stop_annotating


def compute_gradient(J, m, options=None, tape=None, adj_value=1.0):
    """
    Compute the gradient of J with respect to the initialisation value of m,
    that is the value of m at its creation.

    Args:
        J (AdjFloat):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        options (dict): A dictionary of options. To find a list of available options
            have a look at the specific control type.
        tape: The tape to use. Default is the current tape.

    Returns:
        OverloadedType: The derivative with respect to the control. Should be an instance of the same type as
            the control.
    """
    options = options or {}
    tape = tape or get_working_tape()
    tape.reset_variables()
    J.block_variable.adj_value = adj_value
    m = Enlist(m)

    with stop_annotating():
        with tape.marked_nodes(m):
            tape.evaluate_adj(markings=True)

    grads = [i.get_derivative(options=options) for i in m]
    return m.delist(grads)


def compute_hessian(J, m, m_dot, options=None, tape=None):
    """
    Compute the Hessian of J in a direction m_dot at the current value of m

    Args:
        J (AdjFloat):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        m_dot (list or instance of the control type): The direction in which to compute the Hessian.
        options (dict): A dictionary of options. To find a list of available options
            have a look at the specific control type.
        tape: The tape to use. Default is the current tape.

    Returns:
        OverloadedType: The second derivative with respect to the control in direction m_dot. Should be an instance of
            the same type as the control.
    """
    tape = tape or get_working_tape()
    options = options or {}

    tape.reset_tlm_values()
    tape.reset_hessian_values()

    m = Enlist(m)
    m_dot = Enlist(m_dot)
    for i, value in enumerate(m_dot):
        m[i].tlm_value = m_dot[i]

    with stop_annotating():
        tape.evaluate_tlm()

    J.block_variable.hessian_value = 0.0
    with stop_annotating():
        with tape.marked_nodes(m):
            tape.evaluate_hessian(markings=True)

    r = [v.get_hessian(options=options) for v in m]
    return m.delist(r)


def solve_adjoint(J, tape=None, adj_value=1.0):
    """
    Solve the adjoint problem for a functional J.

    This traverses the entire tape backwards, unlike `compute_gradient` which only works out those
    parts of the adjoint necessary to compute the sensitivity with respect to the specified control.
    As a result sensitivities with respect to all intermediate states are accumulated in the
    `adj_value` attribute of the associated block-variables. The adjoint solution of each solution
    step is stored in the `adj_sol` attribute of the corresponding solve block.

    Args:
        J (AdjFloat):  The objective functional.
        tape: The tape to use. Default is the current tape.
    """
    tape = tape or get_working_tape()
    tape.reset_variables()
    J.adj_value = adj_value

    with stop_annotating():
        tape.evaluate_adj(markings=False)



from .drivers import compute_gradient, compute_hessian
from .enlisting import Enlist
from .tape import get_working_tape, stop_annotating, no_annotations
from .overloaded_type import OverloadedType, create_overloaded_object


def _get_extract_derivative_components(derivative_components):
    """
    Construct a function to pass as a pre derivative callback
    when derivative components are required.
    """
    def extract_derivative_components(controls):
        controls_out = Enlist([controls[i]
                               for i in derivative_components])
        return controls_out
    return extract_derivative_components


def _get_pack_derivative_components(controls, derivative_components):
    """
    Construct a function to pass as a post derivative callback
    when derivative components are required.
    """
    def pack_derivative_components(checkpoint, derivatives, values):
        derivatives_out = []
        count = 0
        for i, control in enumerate(controls):
            if i in derivative_components:
                derivatives_out.append(derivatives[count])
                count += 1
            else:
                zero_derivative = control._ad_copy()
                zero_derivative *= 0.
                derivatives_out.append(zero_derivative)
        return derivatives_out
    return pack_derivative_components


class ReducedFunctional(object):
    """Class representing the reduced functional.

    A reduced functional maps a control value to the provided functional.
    It may also be used to compute the derivative of the functional with
    respect to the control.

    Args:
        functional (:obj:`OverloadedType`): An instance of an OverloadedType,
            usually :class:`AdjFloat`. This should be the return value of the
            functional you want to reduce.
        controls (list[Control]): A list of Control instances, which you want
            to map to the functional. It is also possible to supply a single
            Control instance instead of a list.
        derivative_components (tuple of int): The indices of the controls with
            respect to which to take the derivative. By default, the derivative
            is taken with respect to all controls. If present, it overwrites
            derivative_cb_pre and derivative_cb_post.
        derivative_cb_pre (function): Callback function before evaluating
            derivatives. Input is a list of Controls.
            Should return a list of Controls (usually the same
            list as the input) to be passed to compute_gradient.
        derivative_cb_post (function): Callback function after evaluating
            derivatives.  Inputs are: functional.block_variable.checkpoint,
            list of functional derivatives, list of functional values.
            Should return a list of derivatives (usually the same
            list as the input) to be returned from self.derivative.
    """

    def __init__(self, functional, controls,
                 derivative_components=None,
                 scale=1.0, tape=None,
                 eval_cb_pre=lambda *args: None,
                 eval_cb_post=lambda *args: None,
                 derivative_cb_pre=lambda controls: controls,
                 derivative_cb_post=lambda checkpoint, derivative_components, controls: derivative_components,
                 hessian_cb_pre=lambda *args: None,
                 hessian_cb_post=lambda *args: None):
        if not isinstance(functional, OverloadedType):
            raise TypeError("Functional must be an OverloadedType.")
        self.functional = functional
        self.tape = get_working_tape() if tape is None else tape
        self.controls = Enlist(controls)
        self.derivative_components = derivative_components
        self.scale = scale
        self.eval_cb_pre = eval_cb_pre
        self.eval_cb_post = eval_cb_post
        self.derivative_cb_pre = derivative_cb_pre
        self.derivative_cb_post = derivative_cb_post
        self.hessian_cb_pre = hessian_cb_pre
        self.hessian_cb_post = hessian_cb_post

        if self.derivative_components:
            # pre callback
            self.derivative_cb_pre = _get_extract_derivative_components(
                derivative_components)
            # post callback
            self.derivative_cb_post = _get_pack_derivative_components(
                controls, derivative_components)

    def derivative(self, adj_input=1.0, options={}):
        """Returns the derivative of the functional w.r.t. the control.
        Using the adjoint method, the derivative of the functional with
        respect to the control, around the last supplied value of the
        control, is computed and returned.

        Args:
            options (dict): A dictionary of options. To find a list of
                available options have a look at the specific control type.

        Returns:
            OverloadedType: The derivative with respect to the control.
                Should be an instance of the same type as the control.

        """

        # Call callback
        values = [c.tape_value() for c in self.controls]
        controls = self.derivative_cb_pre(self.controls)

        if not controls:
            raise ValueError("""Note that the callback interface
            for derivative_cb_pre has changed. It should now return a
            list of controls (usually the same list as input.""")

        # Scale adjoint input
        with stop_annotating():
            # Make sure `adj_input` is an OverloadedType
            adj_input = create_overloaded_object(adj_input)
            adj_value = adj_input._ad_mul(self.scale)

        derivatives = compute_gradient(self.functional,
                                       controls,
                                       options=options,
                                       tape=self.tape,
                                       adj_value=adj_value)

        # Call callback
        derivatives = self.derivative_cb_post(
            self.functional.block_variable.checkpoint,
            derivatives,
            values)

        if not derivatives:
            raise ValueError("""Note that the callback interface
            for derivative_cb_post has changed. It should now return a
            list of derivatives, usually the same list as input.""")

        return self.controls.delist(derivatives)

    @no_annotations
    def hessian(self, m_dot, options={}):
        """Returns the action of the Hessian of the functional w.r.t. the control on a vector m_dot.

        Using the second-order adjoint method, the action of the Hessian of the
        functional with respect to the control, around the last supplied value
        of the control, is computed and returned.

        Args:
            m_dot ([OverloadedType]): The direction in which to compute the
                action of the Hessian.
            options (dict): A dictionary of options. To find a list of
                available options have a look at the specific control type.

        Returns:
            OverloadedType: The action of the Hessian in the direction m_dot.
                Should be an instance of the same type as the control.
        """
        # Call callback
        values = [c.tape_value() for c in self.controls]
        self.hessian_cb_pre(self.controls.delist(values))

        r = compute_hessian(self.functional, self.controls, m_dot, options=options, tape=self.tape)

        # Call callback
        self.hessian_cb_post(self.functional.block_variable.checkpoint,
                             self.controls.delist(r),
                             self.controls.delist(values))

        return self.controls.delist(r)

    @no_annotations
    def __call__(self, values):
        """Computes the reduced functional with supplied control value.

        Args:
            values ([OverloadedType]): If you have multiple controls this should be a list of
                new values for each control in the order you listed the controls to the constructor.
                If you have a single control it can either be a list or a single object.
                Each new value should have the same type as the corresponding control.

        Returns:
            :obj:`OverloadedType`: The computed value. Typically of instance
                of :class:`AdjFloat`.

        """
        values = Enlist(values)
        if len(values) != len(self.controls):
            raise ValueError("values should be a list of same length as controls.")

        # Call callback.
        self.eval_cb_pre(self.controls.delist(values))

        for i, value in enumerate(values):
            self.controls[i].update(value)

        self.tape.reset_blocks()
        blocks = self.tape.get_blocks()
        with self.marked_controls():
            with stop_annotating():
                for i in self.tape._bar("Evaluating functional").iter(
                    range(len(blocks))
                ):
                    blocks[i].recompute()

        # ReducedFunctional can result in a scalar or an assembled 1-form
        func_value = self.functional.block_variable.saved_output
        # Scale the underlying functional value
        func_value *= self.scale

        # Call callback
        self.eval_cb_post(func_value, self.controls.delist(values))

        return func_value

    def optimize_tape(self):
        self.tape.optimize(
            controls=self.controls,
            functionals=[self.functional]
        )

    def marked_controls(self):
        return marked_controls(self)


class marked_controls(object):
    def __init__(self, rf):
        self.rf = rf

    def __enter__(self):
        for control in self.rf.controls:
            control.mark_as_control()

    def __exit__(self, *args):
        for control in self.rf.controls:
            control.unmark_as_control()



# flake8: noqa

import warnings

from firedrake.adjoint import *

continue_annotation()

warnings.warn("""The firedrake_adjoint module is deprecated.

Instead, use the firedrake.adjoint module and explicitly start taping
by calling continue_annotation().""", FutureWarning)



from itertools import chain
from setuptools import setup

extras = {
    'moola': ['moola>=0.1.6'],
    'test': ['pytest>=3.10', 'flake8', 'coverage'],
    'visualisation': ['tensorflow', 'protobuf',
                      'networkx', 'pygraphviz'],
    'meshing': ['pygmsh', 'meshio'],
    'doc': ['sphinx', 'sphinx-autobuild', 'sphinxcontrib-bibtex']
}
# 'all' includes all of the above
extras['all'] = list(chain(*extras.values()))

setup(name='pyadjoint-ad',
      version='2023.0.0.post0',
      description='High-level automatic differentiation library.',
      author='Jørgen Dokken',
      author_email='dokken@simula.no',
      packages=['firedrake_adjoint',
                'numpy_adjoint',
                'pyadjoint',
                'pyadjoint.optimization'],
      package_dir={'pyadjoint': 'pyadjoint',
                   'firedrake_adjoint': 'firedrake_adjoint',
                   'numpy_adjoint': 'numpy_adjoint'},
      install_requires=['scipy>=1.0'],
      extras_require=extras
      )
